import os
import abc
import warnings
import threading

from typing import Union, Dict, Any, AnyStr


class BaseFileHandler(abc.ABC):
    def __init__(self, path: Union[AnyStr, Any], live_update=False, vault_is_read_only=False):
        self.lock = threading.Lock()
        self.raw_path = path
        self.live_update = live_update
        self.vault_is_read_only = vault_is_read_only
        self.create_resource(path)

    @property
    @abc.abstractmethod
    def resource(self) -> Any:
        """Meant to return the resource that stores the vault in some database."""
        raise NotImplementedError()

    @abc.abstractmethod
    def create_resource(self, path: Union[AnyStr, Any]):
        """Meant to create the resource to store the vault in a database."""
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def path(self) -> Union[AnyStr, Any]:
        """Meant to return the path to the database that stores the vault."""
        raise NotImplementedError()

    @abc.abstractmethod
    def kv_pair_can_be_written(self, obj: Dict) -> bool:
        """Meant to return a bool that says if a given key-value pair can be successfully written to the database."""
        raise NotImplementedError()

    @abc.abstractmethod
    def hash(self) -> str:
        """Meant to return a hash for the database so varvault knows if the vault that is loaded in memory is different from the vault in the file/database."""
        raise NotImplementedError()

    @abc.abstractmethod
    def exists(self) -> bool:
        """Meant to return a bool which says if the resource exists or not"""
        raise NotImplementedError()

    # ================================================================================================================
    # Write
    # ================================================================================================================
    def write(self, vault: dict) -> None:
        f"""Writes the vault to the database by calling the implemented '{self.do_write}' function."""
        if self.vault_is_read_only:
            warnings.warn("Tried to write to a vault-file defined as read-only. This is not permitted by varvault.")
        if self.resource:
            with self.lock:
                self.do_write(vault)

    @abc.abstractmethod
    def do_write(self, vault: dict) -> None:
        """
        A function to write a vault to a file/database. Varvault will call this function internally.
        The class that implements this abstract method has to write a dict to a resource.

        Example:
        `json.dump(vault, open(self.path, "w"), indent=2)`

        :param vault: The vault to write to the file.
        :return: None. Varvault will not use the return value from this function
        """
        raise NotImplementedError()

    # ================================================================================================================
    # Read
    # ================================================================================================================
    def read(self) -> Dict:
        f"""Reads the vault from the database by calling the implemented '{self.do_read}' function."""
        if self.resource:
            with self.lock:
                try:
                    return self.do_read()
                except Exception as e:
                    if self.live_update:
                        # live-update has been defined; This means that any error from reading
                        # the file must be considered okay as the file might not exist yet.
                        return {}
                    raise e
        elif not self.resource and self.live_update:
            try:
                self.create_resource(self.path)
                with self.lock:
                    return self.do_read()
            except (FileNotFoundError, ResourceNotFoundError):
                return {}
        raise ResourceNotFoundError("Tried to read from the resource but the resource isn't created correctly and live-update isn't defined", self)

    @abc.abstractmethod
    def do_read(self) -> Dict:
        """
        A function to read data from a file/database. Varvault will call this function internally.
        The class that implements this abstractmethod has to read data from a file and then return it.

        Example:
        `return json.load(open(self.path))`

        :return: A dict describing the vault from the file.
        """
        raise NotImplementedError()

    def __str__(self):
        return f"resource={self.resource}; path={self.path}; live_update={self.live_update}; vault_is_read_only={self.vault_is_read_only}"


class ResourceNotFoundError(FileNotFoundError):
    def __init__(self, msg, filehandler: BaseFileHandler):
        super(ResourceNotFoundError, self).__init__(msg)
        self.msg = msg
        self.filehandler = filehandler

    def __str__(self):
        return f"{self.msg} - {self.filehandler}"



import os
import json

from typing import Any, Union, Dict, TextIO, AnyStr

from .filehandlers import BaseFileHandler

from .keyring import Key
from .keyring import Keyring

from .validator import validator

from .minivault import MiniVault

from .vault import VarVault

from .vaultflags import VaultFlags

from .vaultfactory import create_vault
from .vaultfactory import from_vault

from .vaultstructs import VaultStructDictBase
from .vaultstructs import VaultStructListBase
from .vaultstructs import VaultStructFloatBase
from .vaultstructs import VaultStructIntBase
from .vaultstructs import VaultStructStringBase

from .utils import concurrently
from .utils import concurrent_execution
from .utils import create_mini_vault_from_file


def clear_logs():
    import glob
    import tempfile

    path = os.path.join(tempfile.tempdir, "varvault-logs", "*.log")
    files = glob.glob(path)
    for f in files:
        os.remove(f)


class JsonFilehandler(BaseFileHandler):

    def __init__(self, path: AnyStr, live_update=False, vault_is_read_only=False):
        path = os.path.expanduser(os.path.expandvars(path))
        self.file_io = None
        super(JsonFilehandler, self).__init__(path, live_update, vault_is_read_only)

    @property
    def resource(self) -> TextIO:
        """Returns the file resource object for this handler."""
        return self.file_io

    def create_resource(self, path: Union[AnyStr, Any]) -> None:
        """Creates the resource self.file_io for this handler which we'll use to read and write to."""
        if path and self.live_update and not os.path.exists(path):
            self.file_io = None
        elif path and not os.path.exists(path):
            # Create the file; It doesn't exist. Try to create the folder first.
            os.makedirs(os.path.dirname(path), exist_ok=True)
            self.file_io = open(path, "w")
            self.file_io.close()
        elif path and os.path.exists(path):
            # The file already exists; Just read from it
            self.file_io = open(path)
            self.file_io.close()
        else:
            raise NotImplementedError("This is not supported")

    @property
    def path(self) -> AnyStr:
        """Returns the path to the vault-file as a JSON file."""
        return self.raw_path

    def kv_pair_can_be_written(self, obj: Dict) -> bool:
        f"""Checks if a key-value pair can be written to a file by attempting to serialize it by using {json.dumps}"""
        try:
            json.dumps(obj)
            return True
        except (TypeError, OverflowError) as e:
            return False

    def hash(self) -> str:
        """Returns the hash for the file, so we can check if the file contains changes compared to what the vault object currently has"""
        import hashlib
        hash_md5 = hashlib.md5()
        with open(self.path, "rb") as f:
            for chunk in iter(lambda: f.read(4096), b""):
                hash_md5.update(chunk)
        return hash_md5.hexdigest()

    def exists(self) -> bool:
        """Returns a bool that determines if the JSON file exists"""
        return os.path.exists(self.path)

    def do_write(self, vault: dict) -> None:
        """Writes the vault to the JSON file"""
        json.dump(vault, open(self.path, "w"), indent=2)

    def do_read(self) -> Dict:
        """Reads the vault from the JSON file"""
        return json.load(open(self.path))
